from eip712_structs import Address, Boolean
from eip712_structs import EIP712Struct as EIP712StructBase
from eip712_structs import Int, Uint, make_domain
from eth_utils import keccak
from hexbytes import HexBytes

from hubble_exchange.constants import (CHAIN_ID, IOCBookContractAddress,
                                       OrderBookContractAddress)
from hubble_exchange.models import IOCOrder as IOCOrderModel
from hubble_exchange.models import LimitOrder as LimitOrderModel

limit_order_domain = make_domain(name='Hubble', version="2.0", chainId=CHAIN_ID, verifyingContract=OrderBookContractAddress)
limit_order_domain_hash = HexBytes(limit_order_domain.hash_struct())

ioc_order_domain = make_domain(name='Hubble', version="2.0", chainId=CHAIN_ID, verifyingContract=IOCBookContractAddress)
ioc_order_domain_hash = HexBytes(ioc_order_domain.hash_struct())


class EIP712Struct(EIP712StructBase):
    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.type_name = cls._name


class LimitOrder(EIP712Struct):
    _name = "Order"

    ammIndex = Uint(256)
    trader = Address()
    baseAssetQuantity = Int(256)
    price = Uint(256)
    salt = Uint(256)
    reduceOnly = Boolean()
    postOnly = Boolean()


class IOCOrder(EIP712Struct):
    _name = "IOCOrder"

    orderType = Uint(8)
    expireAt = Uint(256)
    ammIndex = Uint(256)
    trader = Address()
    baseAssetQuantity = Int(256)
    price = Uint(256)
    salt = Uint(256)
    reduceOnly = Boolean()


def get_limit_order_hash(order: LimitOrderModel) -> HexBytes:
    """
    INCORRECT: use order.get_order_hash() instead
    """
    order_struct = LimitOrder(
        ammIndex=order.amm_index,
        trader=order.trader,
        baseAssetQuantity=order.base_asset_quantity,
        price=order.price,
        salt=order.salt,
        reduceOnly=order.reduce_only,
        postOnly=order.post_only,
    )

    order_struct_hash = HexBytes(order_struct.hash_struct())
    order_hash_bytes = b'\x19\x01' + limit_order_domain_hash + order_struct_hash
    order_hash = HexBytes(keccak(order_hash_bytes))
    return order_hash


def get_ioc_order_hash(order: IOCOrderModel) -> HexBytes:
    order_struct = IOCOrder(
        orderType=1,
        expireAt=order.expire_at,
        ammIndex=order.amm_index,
        trader=order.trader,
        baseAssetQuantity=order.base_asset_quantity,
        price=order.price,
        salt=order.salt,
        reduceOnly=order.reduce_only,
    )

    order_struct_hash = HexBytes(order_struct.hash_struct())
    order_hash_bytes = b'\x19\x01' + ioc_order_domain_hash + order_struct_hash
    order_hash = HexBytes(keccak(order_hash_bytes))
    return order_hash



import json
from typing import Dict, List

import websockets
from hexbytes import HexBytes

from hubble_exchange.eip712 import get_ioc_order_hash
from hubble_exchange.eth import get_async_web3_client, get_websocket_endpoint
from hubble_exchange.models import (AsyncOrderBookDepthCallback,
                                    AsyncOrderStatusCallback,
                                    AsyncPlaceOrdersCallback,
                                    AsyncPositionCallback,
                                    AsyncSubscribeToOrderBookDepthCallback,
                                    ConfirmationMode, IOCOrder, LimitOrder,
                                    MarketFeedUpdate,
                                    OrderBookDepthUpdateResponse,
                                    OrderStatusResponse, TraderFeedUpdate,
                                    WebsocketResponse)
from hubble_exchange.order_book import OrderBookClient, TransactionMode
from hubble_exchange.utils import (add_0x_prefix, float_to_scaled_int,
                                   get_address_from_private_key, get_new_salt)


class HubbleClient:
    def __init__(self, private_key: str):
        if not private_key:
            raise ValueError("Private key is not set")
        self.trader_address = get_address_from_private_key(private_key)
        if not self.trader_address:
            raise ValueError("Cannot determine trader address from private key")

        self.web3_client = get_async_web3_client()
        self.websocket_endpoint = get_websocket_endpoint()
        self.order_book_client = OrderBookClient(private_key)

    def set_transaction_mode(self, mode: TransactionMode):
        self.order_book_client.set_transaction_mode(mode)

    async def get_nonce(self):
        return await self.order_book_client.get_current_nonce()

    async def get_markets(self):
        return await self.order_book_client.get_markets()

    async def get_order_book(self, market: int, callback: AsyncOrderBookDepthCallback):
        order_book_depth = await self.web3_client.eth.get_order_book_depth(market)
        return await callback(order_book_depth)

    async def get_margin_and_positions(self, callback: AsyncPositionCallback):
        response = await self.web3_client.eth.get_margin_and_positions(self.trader_address)
        return await callback(response)

    async def get_limit_order_details(self, order_id: HexBytes, callback: AsyncOrderStatusCallback):
        """ Works only for open limit orders """
        response = await self.web3_client.eth.get_order_status(order_id.hex())
        return await callback(response)

    async def get_limit_order_status(self, order_id: HexBytes, callback):
        response = await self.order_book_client.get_limit_order_status(order_id)
        return await callback(response)

    async def get_open_orders(self, market_id: int, callback: AsyncOrderStatusCallback):
        response = await self.web3_client.eth.get_open_orders(self.trader_address, market_id)
        return await callback(response)
    
    async def get_trades(self, market, start_time, end_time, callback):
        if start_time is None or end_time is None:
            raise ValueError("Start and end time must be specified")

        response = await self.order_book_client.get_trades(self.trader_address, market, start_time, end_time)
        return await callback(response)
    
    async def place_limit_orders(self, orders: List[LimitOrder], wait_for_response: bool, callback: AsyncPlaceOrdersCallback, tx_options = None, mode=None):
        if len(orders) > 75:
            raise ValueError("Cannot place more than 75 orders at once")

        for order in orders:
            if order.amm_index is None:
                raise ValueError("Order AMM index is not set")
            if order.base_asset_quantity is None:
                raise ValueError("Order base asset quantity is not set")
            if order.price is None:
                raise ValueError("Order price is not set")
            if order.reduce_only is None:
                raise ValueError("Order reduce only is not set")
            if order.post_only is None:
                raise ValueError("Order post only is not set")

            # trader and salt can be set automatically
            if order.trader in [None, "0x", ""]:
                order.trader = self.trader_address
            if order.salt in [None, 0]:
                order.salt = get_new_salt()

        order_ids = set()  # stores all the order ids
        for order in orders:
            order_hash = order.get_order_hash()
            order.id = order_hash
            order_ids.add(order_hash)  # add each order id to the set

        # if the response if requested then we'll have to wait for the transaction to be mined
        # This is because the receipt is generated only after the transaction is mined(accepted)
        if wait_for_response:
            mode = TransactionMode.wait_for_accept

        tx_hash = await self.order_book_client.place_limit_orders(orders, tx_options, mode)
        if wait_for_response:
            receipt = await self.web3_client.eth.get_transaction_receipt(tx_hash)

            event_order_ids = set()  # stores order ids present in events
            response = []
            accepted_events = self.order_book_client.get_events_from_receipt(receipt, "OrderAccepted")
            for event in accepted_events:
                order_id = event.args.orderHash
                event_order_ids.add(add_0x_prefix(order_id.hex()))
                response.append({
                    "order_id": add_0x_prefix(order_id.hex()),
                    "success": True
                })

            rejected_events = self.order_book_client.get_events_from_receipt(receipt, "OrderRejected")
            for event in rejected_events:
                order_id = event.args.orderHash
                event_order_ids.add(add_0x_prefix(order_id.hex()))
                response.append({
                    "order_id": add_0x_prefix(order_id.hex()),
                    "success": False,
                    "error": event.args.err
                })

            # check for missing order ids; should never happen
            missing_order_ids = order_ids - event_order_ids
            for missing_id in missing_order_ids:
                response.append({
                    "order_id": missing_id,
                    "success": False,
                    "error": "Unknown error"
                })

            return await callback(response)
        else:
            return await callback(orders)

    async def place_ioc_orders(self, orders: List[IOCOrder], wait_for_response: bool, callback: AsyncPlaceOrdersCallback, tx_options = None, mode=None):
        if len(orders) > 75:
            raise ValueError("Cannot place more than 75 orders at once")

        for order in orders:
            if order.amm_index is None:
                raise ValueError("Order AMM index is not set")
            if order.base_asset_quantity is None:
                raise ValueError("Order base asset quantity is not set")
            if order.price is None:
                raise ValueError("Order price is not set")
            if order.reduce_only is None:
                raise ValueError("Order reduce only is not set")
            if order.expire_at is None:
                raise ValueError("Order expiry is not set")

            # trader and salt can be set automatically
            if order.trader in [None, "0x", ""]:
                order.trader = self.trader_address
            if order.salt in [None, 0]:
                order.salt = get_new_salt()

        order_ids = set()  # stores all the order ids
        for order in orders:
            order_hash = get_ioc_order_hash(order)
            order.id = order_hash
            order_ids.add(order_hash)  # add each order id to the set

        # if the response if requested then we'll have to wait for the transaction to be mined
        # This is because the receipt is generated only after the transaction is mined(accepted)
        if wait_for_response:
            mode = TransactionMode.wait_for_accept

        tx_hash = await self.order_book_client.place_ioc_orders(orders, tx_options, mode)
        if wait_for_response:
            receipt = await self.web3_client.eth.get_transaction_receipt(tx_hash)

            events = self.order_book_client.get_events_from_receipt(receipt, "OrderPlaced", contract_name="ioc")
            event_order_ids = set()  # stores order ids present in events
            response = []
            for event in events:
                order_id = event.args.orderHash
                event_order_ids.add(add_0x_prefix(order_id.hex()))
                response.append({
                    "order_id": add_0x_prefix(order_id.hex()),
                    "success": True
                })

            missing_order_ids = order_ids - event_order_ids  # order ids not present in events

            for missing_id in missing_order_ids:
                response.append({
                    "order_id": missing_id,
                    "success": False
                })

            return await callback(response)
        else:
            return await callback(orders)

    async def cancel_limit_orders(self, orders: List[LimitOrder], wait_for_response: bool, callback, tx_options = None, mode=None):
        if len(orders) > 100:
            raise ValueError("Cannot cancel more than 100 orders at once")

        # if the response if requested then we'll have to wait for the transaction to be mined
        # This is because the receipt is generated only after the transaction is mined(accepted)
        if wait_for_response:
            mode = TransactionMode.wait_for_accept

        order_ids = set()  # stores all the order ids
        for order in orders:
            order_hash = order.get_order_hash()
            order.id = order_hash
            order_ids.add(order_hash)  # add each order id to the set

        tx_hash = await self.order_book_client.cancel_orders(orders, tx_options, mode)

        if wait_for_response:
            receipt = await self.web3_client.eth.get_transaction_receipt(tx_hash)

            response = []
            event_order_ids = set()
            cancelled_events = self.order_book_client.get_events_from_receipt(receipt, "OrderCancelAccepted")
            for event in cancelled_events:
                event_order_ids.add(add_0x_prefix(event.args.orderHash.hex()))
                response.append({
                    "order_id": add_0x_prefix(event.args.orderHash.hex()),
                    "success": True
                })
            rejected_events = self.order_book_client.get_events_from_receipt(receipt, "OrderCancelRejected")
            for event in rejected_events:
                event_order_ids.add(add_0x_prefix(event.args.orderHash.hex()))
                response.append({
                    "order_id": add_0x_prefix(event.args.orderHash.hex()),
                    "success": False,
                    "error": event.args.err
                })

            missing_order_ids = order_ids - event_order_ids  # order ids not present in events
            for missing_id in missing_order_ids:
                response.append({
                    "order_id": missing_id,
                    "success": False,
                    "error": "Unknown error"
                })

            return await callback(response)
        else:
            return await callback(orders)

    async def cancel_order_by_id(self, order_id: HexBytes, wait_for_response: bool, callback, tx_options = None, mode=None):
        async def order_status_callback(response: OrderStatusResponse) -> LimitOrder:
            position_side_multiplier = 1 if response.positionSide == "LONG" else -1
            return LimitOrder(
                id=order_id,
                amm_index=response.symbol,
                trader=self.trader_address,
                base_asset_quantity=float_to_scaled_int(float(response.origQty) * position_side_multiplier, 18),
                price=float_to_scaled_int(float(response.price), 6),
                salt=int(response.salt),
                reduce_only=response.reduceOnly,
                post_only=response.postOnly,
            )
        order = await self.get_limit_order_details(order_id, order_status_callback)
        return await self.cancel_limit_orders([order], True, wait_for_response, callback, tx_options, mode)

    async def get_order_fills(self, order_id: str) -> List[Dict]:
        return await self.order_book_client.get_order_fills(order_id)

    async def subscribe_to_order_book_depth(
        self, market: int, callback: AsyncSubscribeToOrderBookDepthCallback
    ) -> None:
        async with websockets.connect(self.websocket_endpoint) as ws:
            msg = {
                "jsonrpc": "2.0",
                "id": 1,
                "method": "trading_subscribe",
                "params": ["streamDepthUpdateForMarket", market]
            }
            await ws.send(json.dumps(msg))

            async for message in ws:
                message_json = json.loads(message)
                if message_json.get('result'):
                    # ignore because it's a subscription confirmation with subscription id
                    continue
                response = WebsocketResponse(**message_json)
                if response.method and response.method == "trading_subscription":
                    response = OrderBookDepthUpdateResponse(
                        T=response.params['result']['T'],
                        symbol=response.params['result']['s'],
                        bids=response.params['result']['b'],
                        asks=response.params['result']['a'],
                    )
                    await callback(ws, response)

    async def subscribe_to_trader_updates(
        self, update_type: ConfirmationMode, callback
    ) -> None:
        async with websockets.connect(self.websocket_endpoint) as ws:
            msg = {
                "jsonrpc": "2.0",
                "id": 1,
                "method": "trading_subscribe",
                "params": ["streamTraderUpdates", self.trader_address, update_type.value]
            }
            await ws.send(json.dumps(msg))

            async for message in ws:
                message_json = json.loads(message)
                if message_json.get('result'):
                    # ignore because it's a subscription confirmation with subscription id
                    continue
                response = WebsocketResponse(**message_json)
                if response.method and response.method == "trading_subscription":
                    response = TraderFeedUpdate(**response.params['result'])
                    await callback(ws, response)

    async def subscribe_to_market_updates(
        self, market, update_type: ConfirmationMode, callback
    ) -> None:
        async with websockets.connect(self.websocket_endpoint) as ws:
            msg = {
                "jsonrpc": "2.0",
                "id": 1,
                "method": "trading_subscribe",
                "params": ["streamMarketTrades", market, update_type.value]
            }
            await ws.send(json.dumps(msg))

            async for message in ws:
                message_json = json.loads(message)
                if message_json.get('result'):
                    # ignore because it's a subscription confirmation with subscription id
                    continue
                response = WebsocketResponse(**message_json)
                if response.method and response.method == "trading_subscription":
                    response = MarketFeedUpdate(**response.params['result'])
                    await callback(ws, response)



import math
import random
import time

from eth_account import Account
from eth_typing import Address
from eth_utils.hexadecimal import add_0x_prefix as add_0x_prefix_eth_utils
from hexbytes import HexBytes


def float_to_scaled_int(val: float, scale: int) -> int:
    big_float = math.floor(val * (10 ** scale))
    return int(big_float)


def int_to_scaled_float(val: int, scale: int) -> float:
    return float(val) / (10 ** scale)


def get_address_from_private_key(private_key: str) -> Address:
    account = Account.from_key(private_key)
    return account.address

def get_new_salt() -> int:
    return int(str(time.time_ns()) + str(random.randint(0, 10000)))


def add_0x_prefix(value: str) -> HexBytes:
    return HexBytes(add_0x_prefix_eth_utils(value))



import json
import os
from typing import Any, Dict, List

from eth_typing import Address
from hexbytes import HexBytes
from web3.contract.async_contract import AsyncContractFunction
from web3.logs import DISCARD

from hubble_exchange.constants import (CHAIN_ID, GAS_PER_ORDER, MAX_GAS_LIMIT,
                                       ClearingHouseContractAddress,
                                       IOCBookContractAddress,
                                       OrderBookContractAddress)
from hubble_exchange.eth import get_async_web3_client, get_sync_web3_client
from hubble_exchange.models import ExecutionMode, IOCOrder, LimitOrder, OrderStatus, Trade, TransactionMode
from hubble_exchange.utils import (get_address_from_private_key,
                                   int_to_scaled_float)

# read abi from file
HERE = os.path.dirname(__file__)
with open(f"{HERE}/contract_abis/OrderBook.json", 'r') as abi_file:
    abi_str = abi_file.read()
    ORDERBOOK_ABI = json.loads(abi_str)

with open(f"{HERE}/contract_abis/ImmediayeOrCancelOrders.json", 'r') as abi_file:
    abi_str = abi_file.read()
    IOC_ORDERBOOK_ABI = json.loads(abi_str)

with open(f"{HERE}/contract_abis/ClearingHouse.json", 'r') as abi_file:
    abi_str = abi_file.read()
    CLEARINGHOUSE_ABI = json.loads(abi_str)

with open(f"{HERE}/contract_abis/AMM.json", 'r') as abi_file:
    abi_str = abi_file.read()
    AMM_ABI = json.loads(abi_str)


class OrderBookClient(object):
    def __init__(self, private_key: str):
        self._private_key = private_key
        self.public_address = get_address_from_private_key(private_key)

        self.web3_client = get_async_web3_client()
        self.order_book_contract = self.web3_client.eth.contract(address=OrderBookContractAddress, abi=ORDERBOOK_ABI)
        self.ioc_order_book_contract = self.web3_client.eth.contract(address=IOCBookContractAddress, abi=IOC_ORDERBOOK_ABI)
        self.clearing_house_contract = self.web3_client.eth.contract(address=ClearingHouseContractAddress, abi=CLEARINGHOUSE_ABI)

        # get nonce from sync web3 client
        sync_web3 = get_sync_web3_client()
        self.nonce = sync_web3.eth.get_transaction_count(self.public_address)

        self.transaction_mode = TransactionMode.no_wait  # default

    def set_transaction_mode(self, mode: TransactionMode):
        self.transaction_mode = mode

    async def get_markets(self):
        amm_addresses = await self.clearing_house_contract.functions.getAMMs().call()

        markets = {}
        for i, amm_address in enumerate(amm_addresses):
            amm_contract = self.web3_client.eth.contract(address=amm_address, abi=AMM_ABI)
            name = await amm_contract.functions.name().call()
            markets[i] = name

        return markets

    async def get_limit_order_status(self, order_id: HexBytes):
        response = await self.order_book_contract.functions.orderStatus(order_id).call()
        status = response[3]
        filled_amount = response[1]
        if status == OrderStatus.Placed.value and filled_amount > 0:
            status = OrderStatus.PartiallyFilled.value
        return OrderStatus(status).name

    async def place_limit_orders(self, orders: List[LimitOrder], custom_tx_options=None, mode=None):
        """
        Place limit orders
        """
        place_order_payload = []

        for order in orders:
            place_order_payload.append(order.to_dict())

        tx_options = {'gas': min(GAS_PER_ORDER * len(orders), MAX_GAS_LIMIT)}
        tx_options.update(custom_tx_options or {})
        method = getattr(self.order_book_contract.functions, "placeOrders")
        return await self._send_orderbook_transaction(method, [place_order_payload], tx_options, mode)

    async def place_ioc_orders(self, orders: List[IOCOrder], custom_tx_options=None, mode=None):
        """
        Place ioc orders
        """
        place_order_payload = []

        for order in orders:
            place_order_payload.append(order.to_dict())

        tx_options = {'gas': min(GAS_PER_ORDER * len(orders), MAX_GAS_LIMIT)}
        tx_options.update(custom_tx_options or {})
        method = getattr(self.ioc_order_book_contract.functions, "placeOrders")
        return await self._send_orderbook_transaction(method, [place_order_payload], tx_options, mode)

    async def cancel_orders(self, orders: list[LimitOrder], custom_tx_options=None, mode=None):
        cancel_order_payload = []
        for order in orders:
            cancel_order_payload.append(order.to_dict())

        tx_options = {'gas': min(GAS_PER_ORDER * len(orders), MAX_GAS_LIMIT)}
        tx_options.update(custom_tx_options or {})

        method = getattr(self.order_book_contract.functions, "cancelOrders")
        return await self._send_orderbook_transaction(method, [cancel_order_payload], tx_options, mode)

    async def get_order_fills(self, order_id: str) -> List[Dict]:
        orders_matched_events = await self.order_book_contract.events.OrderMatched().get_logs(
            {"orderHash": order_id},
            fromBlock='earliest',
        )

        fills = []
        for event in orders_matched_events:
            fills.append({
                "block_number": event.blockNumber,
                "transaction_hash": event.transactionHash,
                "timestamp": event.args.timestamp,
                "fill_amount": int_to_scaled_float(event.args.fillAmount, 18),
                "price": int_to_scaled_float(event.args.price, 6),
            })
        return fills

    async def get_trades(self, trader:Address, market:int, start_time:int, end_time:int) -> List[Dict]:
        query = {"trader": trader}
        if market is not None:
            query["idx"] = market

        latest_block_number = await self.web3_client.eth.get_block_number()

        # iterate in chunks till event.timestamp <= start_time
        CHUNK_SIZE = 20000
        trades = []
        to_block = latest_block_number
        while True:
            if to_block <= 0:
                break
            from_block = max(to_block - CHUNK_SIZE - 1, 0)  # because from and to blocks both are inclusive
            trades_events = await self.clearing_house_contract.events.PositionModified().get_logs(
                query,
                fromBlock=from_block,
                toBlock=to_block,
            )

            # break early if even the first event is after end_time
            if len(trades_events) == 0 or trades_events[0].args.timestamp > end_time:
                # get the earliest block and check its timestamp; exit early if it's less than start_time
                earliest_block = await self.web3_client.eth.get_block(from_block)
                if earliest_block.timestamp < start_time:
                    break

                to_block -= CHUNK_SIZE
                continue

            # reversed because we want the response to be from latest to earliest
            for event in reversed(trades_events):
                if event.args.timestamp < start_time:
                    break
                if event.args.timestamp <= end_time:
                    trades.append(Trade(
                        BlockNumber=event.blockNumber,
                        TransactionHash=event.transactionHash,
                        Market=event.args.idx,
                        Timestamp=event.args.timestamp,
                        TradedAmount=int_to_scaled_float(event.args.baseAsset, 18),
                        Price=int_to_scaled_float(event.args.price, 6),
                        RealizedPnl=int_to_scaled_float(event.args.realizedPnl, 6),
                        OpenNotional=int_to_scaled_float(event.args.openNotional, 18),
                        ExecutionMode=ExecutionMode(event.args.mode).name,
                    ))

            to_block -= CHUNK_SIZE
            continue

        return trades

    def get_events_from_receipt(self, receipt, event_name, contract_name=None):
        if contract_name is None:
            contract = self.order_book_contract
        if contract_name == "ioc":
            contract = self.ioc_order_book_contract
        event = getattr(contract.events, event_name)
        return event().process_receipt(receipt, DISCARD)

    async def get_current_nonce(self) -> int:
        return await self.web3_client.eth.get_transaction_count(self.public_address)

    async def _get_nonce(self) -> int:
        if self.nonce is None:
            self.nonce = await self.web3_client.eth.get_transaction_count(self.public_address)
        else:
            self.nonce += 1
        return self.nonce - 1

    async def _send_orderbook_transaction(self, method: AsyncContractFunction, args: List[Any], tx_options: Dict, mode: TransactionMode) -> HexBytes:
        if mode is None:
            mode = self.transaction_mode

        nonce = await self._get_nonce()
        tx_params = {
            'from': self.public_address,
            'chainId': CHAIN_ID,
            'nonce': nonce,
            # "maxFeePerGas": web3.Web3.to_wei(90, 'gwei'),
            # "maxPriorityFeePerGas": web3.Web3.to_wei(30, 'gwei'),
        }
        if tx_options:
            tx_params.update(tx_options)

        transaction = await method(*args).build_transaction(tx_params)
        signed_tx = self.web3_client.eth.account.sign_transaction(transaction, self._private_key)
        tx_hash = await self.web3_client.eth.send_raw_transaction(signed_tx.rawTransaction)
        if mode == TransactionMode.wait_for_accept:
            await self.web3_client.eth.wait_for_transaction_receipt(tx_hash, timeout=120, poll_latency=0.1)
        elif mode == TransactionMode.wait_for_head:
            await self.web3_client.eth.wait_for_transaction_status(tx_hash, timeout=120, poll_latency=0.1)

        return tx_hash




import asyncio
import os
from typing import Awaitable, Callable, Dict, List, Optional

from eth_typing import Address
from hexbytes import HexBytes
from web3 import AsyncHTTPProvider, AsyncWeb3, HTTPProvider, Web3
from web3.eth.async_eth import AsyncEth
from web3.exceptions import TimeExhausted
from web3.main import Web3, get_async_default_modules
from web3.method import Method, default_root_munger
from web3.middleware import (async_geth_poa_middleware,
                             construct_simple_cache_middleware,
                             geth_poa_middleware)
from web3.middleware.async_cache import _async_simple_cache_middleware
from web3.types import RPCEndpoint, TxParams, Wei, _Hash32

from hubble_exchange.errors import OrderNotFound, TraderNotFound
from hubble_exchange.models import (GetPositionsResponse, OpenOrder,
                                    OrderBookDepthResponse,
                                    OrderStatusResponse)

sync_web3_client = None
async_web3_client = None


class HubblenetEth(AsyncEth):
    _get_transaction_status: Method[Callable[[_Hash32], Awaitable[Dict]]] = Method(RPCEndpoint("eth_getTransactionStatus"), mungers=[default_root_munger])
    _get_order_status: Method[Callable[[_Hash32], Awaitable[Dict]]] = Method(RPCEndpoint("trading_getOrderStatus"), mungers=[default_root_munger])
    _get_margin_and_positions: Method[Callable[[Address], Awaitable[Dict]]] = Method(RPCEndpoint("trading_getMarginAndPositions"), mungers=[default_root_munger])
    _get_order_book_depth: Method[Callable[[int], Awaitable[Dict]]] = Method(RPCEndpoint("trading_getTradingOrderBookDepth"), mungers=[default_root_munger])
    _get_open_orders: Method[Callable[[Address, str], Awaitable[Dict]]] = Method(RPCEndpoint("orderbook_getOpenOrders"), mungers=[default_root_munger])

    async def get_order_status(self, order_id: _Hash32) -> OrderStatusResponse:
        try:
            response = await self._get_order_status(order_id)
            return OrderStatusResponse(**response)
        except ValueError as e:
            if len(e.args) > 0 and e.args[0].get('message', '') == "order not found":
                raise OrderNotFound()
            else:
                raise e

    async def get_margin_and_positions(self, trader: Address) -> GetPositionsResponse:
        try:
            response = await self._get_margin_and_positions(trader)
            return GetPositionsResponse(**response)
        except ValueError as e:
            if len(e.args) > 0 and e.args[0].get('message', '') == "trader not found":
                raise TraderNotFound()
            else:
                raise e

    async def get_order_book_depth(self, market: int) -> OrderBookDepthResponse:
        response = await self._get_order_book_depth(market)
        return OrderBookDepthResponse(**response)

    async def get_open_orders(self, trader:Address, market: int=None) -> List[OpenOrder]:
        if market is None:
            market_str = ""
        else:
            market_str = str(market)

        open_orders = []    
        response = await self._get_open_orders(trader, market_str)
        for order in response["Orders"]:
            open_orders.append(OpenOrder(**order))
        return open_orders

    async def wait_for_transaction_status(self, transaction_hash: HexBytes, timeout: float = 120, poll_latency: float = 0.1) -> Dict:
        async def _wait_for_status_with_timeout(
            _tx_hash: _Hash32, _poll_latency: float
        ) -> Dict:
            while True:
                try:
                    tx_status = await self._get_transaction_status(_tx_hash.hex())
                except:
                    tx_status = None
                if tx_status is None or tx_status['status'] == "NOT_FOUND":
                    await asyncio.sleep(_poll_latency)
                    continue
                else:
                    break
            return tx_status

        try:
            return await asyncio.wait_for(
                _wait_for_status_with_timeout(transaction_hash, poll_latency),
                timeout=timeout,
            )
        except asyncio.TimeoutError:
            raise TimeExhausted(
                f"Transaction {HexBytes(transaction_hash) !r} is not in the chain "
                f"after {timeout} seconds"
            )


class HubblenetWeb3(AsyncWeb3):
    eth: HubblenetEth


def get_web3_modules() -> Dict:
    modules = get_async_default_modules()
    modules["eth"] = HubblenetEth
    return modules


def get_async_web3_client() -> HubblenetWeb3:
    global async_web3_client
    if not async_web3_client:
        rpc_endpoint = get_rpc_endpoint()

        async_web3_client = HubblenetWeb3(AsyncHTTPProvider(rpc_endpoint), modules=get_web3_modules())
        async_web3_client.middleware_onion.inject(async_geth_poa_middleware, layer=0)

        # cache frequent eth_chainId calls
        async_web3_client.middleware_onion.add(_async_simple_cache_middleware, name="cache")

        # async_web3_client.eth.set_gas_price_strategy(rpc_gas_price_strategy)

    return async_web3_client


def get_sync_web3_client() -> Web3:
    global sync_web3_client
    if not sync_web3_client:
        rpc_endpoint = get_rpc_endpoint()

        sync_web3_client = Web3(HTTPProvider(rpc_endpoint))
        sync_web3_client.middleware_onion.inject(geth_poa_middleware, layer=0)

        sync_web3_client.middleware_onion.add(construct_simple_cache_middleware(), name="cache")

    return sync_web3_client


def get_rpc_endpoint() -> str:
    rpc_host = os.getenv("HUBBLE_RPC_HOST")
    if not rpc_host:
        raise ValueError("HUBBLE_RPC_HOST environment variable not set")
    blockchain_id = os.getenv("HUBBLE_BLOCKCHAIN_ID")
    if not blockchain_id:
        raise ValueError("HUBBLE_BLOCKCHAIN_ID environment variable not set")
    path = f"/ext/bc/{blockchain_id}/rpc"
    rpc_endpoint = f"https://{rpc_host}{path}"
    return rpc_endpoint


def get_websocket_endpoint() -> str:
    rpc_host = os.getenv("HUBBLE_RPC_HOST")
    if not rpc_host:
        raise ValueError("HUBBLE_RPC_HOST environment variable not set")
    blockchain_id = os.getenv("HUBBLE_BLOCKCHAIN_ID")
    if not blockchain_id:
        raise ValueError("HUBBLE_BLOCKCHAIN_ID environment variable not set")
    path = f"/ext/bc/{blockchain_id}/ws"
    websocket_endpoint = f"wss://{rpc_host}{path}"
    return websocket_endpoint



from hubble_exchange.client import HubbleClient
from hubble_exchange.models import (ConfirmationMode, GetPositionsResponse,
                                    IOCOrder, LimitOrder,
                                    OrderBookDepthResponse,
                                    OrderBookDepthUpdateResponse,
                                    OrderStatusResponse)
from hubble_exchange.order_book import TransactionMode



import time
from dataclasses import dataclass
from enum import Enum
from typing import Any, Coroutine, Dict, List

from eth_typing import Address
from hexbytes import HexBytes
from typing_extensions import Protocol
from web3 import Web3

from hubble_exchange.utils import float_to_scaled_int, get_new_salt


@dataclass
class LimitOrder:
    id: HexBytes
    amm_index: int
    trader: Address
    base_asset_quantity: int
    price: int
    salt: int
    reduce_only: bool
    post_only: bool

    def to_dict(self):
        return {
            "ammIndex": self.amm_index,
            "trader": Web3.to_checksum_address(self.trader),
            "baseAssetQuantity": self.base_asset_quantity,
            "price": self.price,
            "salt": self.salt,
            "reduceOnly": self.reduce_only,
            "postOnly": self.post_only,
        }

    @classmethod
    def new(cls, amm_index: int, base_asset_quantity: float, price: float, reduce_only: bool, post_only: bool):
        """
        Create a new order with a random salt and no ID or trader. This can be used for placing
        multiple orders at once.
        """
        return cls(
            id=None, # type: ignore
            amm_index=amm_index,
            trader=None, # type: ignore
            base_asset_quantity=float_to_scaled_int(base_asset_quantity, 18),
            price=float_to_scaled_int(price, 6),
            salt=get_new_salt(),
            reduce_only=reduce_only,
            post_only=post_only)

    def get_order_hash(self):
        """
        Implements the same logic as the contract's getOrderHashV2 function  - keccak256(abi.encode(order))
        It's converting each field to bytes and then concatenating them together instead of using Web3.solidity_keccak function
        because it was giving incorrect results.
        One important part is to pad all the fields to 32 bytes
        """

        def int_to_bytes(value: int):
            return value.to_bytes(32, 'big', signed=True)

        packed_data = (
            int_to_bytes(self.amm_index) +
            Web3.to_bytes(hexstr=self.trader).rjust(32, b'\0') +
            int_to_bytes(self.base_asset_quantity) +
            int_to_bytes(self.price) +
            int_to_bytes(self.salt) +
            (b'\x01' if self.reduce_only else b'\x00').rjust(32, b'\0') +
            (b'\x01' if self.post_only else b'\x00').rjust(32, b'\0')
        )
        return Web3.keccak(packed_data)


@dataclass
class IOCOrder:
    id: HexBytes
    amm_index: int
    trader: Address
    base_asset_quantity: int
    price: int
    salt: int
    reduce_only: bool
    expire_at: int

    def to_dict(self):
        return {
            "ammIndex": self.amm_index,
            "trader": self.trader,
            "baseAssetQuantity": self.base_asset_quantity,
            "price": self.price,
            "salt": self.salt,
            "reduceOnly": self.reduce_only,
            "expireAt": self.expire_at,
            "orderType": 1,  # hardcoded type = 1 for ioc orders
        }

    @classmethod
    def new(cls, amm_index: int, base_asset_quantity: float, price: float, reduce_only: bool, expiry_duration: int):
        """
        Create a new order with a random salt and no ID or trader. This can be used for placing
        multiple orders at once.
        """
        return cls(
            id=None, # type: ignore
            amm_index=amm_index,
            trader=None, # type: ignore
            base_asset_quantity=float_to_scaled_int(base_asset_quantity, 18),
            price=float_to_scaled_int(price, 6),
            salt=get_new_salt(),
            reduce_only=reduce_only,
            expire_at=int(time.time()) + expiry_duration)


@dataclass
class OrderStatusResponse:
    executedQty: str
    orderId: str
    origQty: str
    price: str
    reduceOnly: bool
    postOnly: bool
    positionSide: str
    status: str
    symbol: int
    time: int
    type: str
    updateTime: int
    salt: int


@dataclass
class OrderBookDepthResponse:
    lastUpdateId: int
    E: int
    T: int
    symbol: int
    bids: List[List[str]]
    asks: List[List[str]]


class Market(int):
    pass


@dataclass
class Position:
    market: Market
    openNotional: str
    size: str
    unrealisedFunding: str
    liquidationThreshold: str
    notionalPosition: str
    unrealisedProfit: str
    marginFraction: str
    liquidationPrice: str
    markPrice: str


@dataclass
class GetPositionsResponse:
    margin: str
    reservedMargin: str
    positions: List[Position]


@dataclass
class Message:
    jsonrpc: str
    id: int
    method: str
    params: List[Any]


@dataclass
class Params:
    subscription: str
    result: Any


@dataclass
class WebsocketResponse:
    jsonrpc: str
    method: str
    params: Params


@dataclass
class OrderBookDepthUpdateResponse:
    T: int
    symbol: int
    bids: List[List[str]]
    asks: List[List[str]]


@dataclass
class OpenOrder:
    Market: int
    Price: str
    Size: str
    FilledSize: str
    Timestamp: int
    Salt: str
    OrderId: str
    ReduceOnly: bool
    PostOnly: bool
    OrderType: str


@dataclass
class Trade:
    BlockNumber: int
    TransactionHash: HexBytes
    Market: int
    Timestamp: int
    TradedAmount: float
    Price: float
    RealizedPnl: float
    OpenNotional: float
    ExecutionMode: str


@dataclass
class TraderFeedUpdate:
    Trader: Address
    OrderId: HexBytes
    OrderType: str
    Removed: bool
    EventName: str
    Args: Dict
    BlockNumber: int
    BlockStatus: str
    Timestamp: int
    TransactionHash: HexBytes


@dataclass
class MarketFeedUpdate:
    Trader: Address
    Market: int
    Size: float
    Price: float
    Removed: bool
    EventName: str
    BlockNumber: int
    BlockStatus: str
    Timestamp: int
    TransactionHash: HexBytes


class TransactionMode(Enum):
    no_wait = 0
    wait_for_head = 1
    wait_for_accept = 2


class ExecutionMode(Enum):
    Taker = 0
    Maker = 1
    SameBlock = 2  # not used
    Liquidation = 3


class OrderStatus(Enum):
    Invalid = 0
    Placed = 1
    Filled = 2
    Cancelled = 3
    PartiallyFilled = 4  # this status is for sdk only; not present in contract


class AsyncOrderBookDepthCallback(Protocol):
    def __call__(self, response: OrderBookDepthResponse) -> Coroutine[Any, Any, Any]: ...


class AsyncPositionCallback(Protocol):
    def __call__(self, response: GetPositionsResponse) -> Coroutine[Any, Any, Any]: ...


class AsyncOrderStatusCallback(Protocol):
    def __call__(self, response: OrderStatusResponse) -> Coroutine[Any, Any, Any]: ...


class AsyncPlaceOrdersCallback(Protocol):
    def __call__(self, response: List[Any]) -> Coroutine[Any, Any, Any]: ...


class AsyncSubscribeToOrderBookDepthCallback(Protocol):
    def __call__(self, ws, response: OrderBookDepthUpdateResponse) -> Coroutine[Any, Any, Any]: ...


class ConfirmationMode(Enum):
    head = "head"
    accepted = "accepted"



from eth_typing import Address

OrderBookContractAddress = Address("0x0300000000000000000000000000000000000000")
IOCBookContractAddress = Address("0x635c5F96989a4226953FE6361f12B96c5d50289b")
ClearingHouseContractAddress = Address("0x0300000000000000000000000000000000000002")

CHAIN_ID = 321123
MAX_GAS_LIMIT = 7_000_000  # 7 million
GAS_PER_ORDER = 300_000  # 300k
